@@ -12,20 +12,24 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import sys, os
+import os
 import subprocess
 import shutil
 import tempfile
+from ..environment import detect_ninja, detect_scanbuild
+from ..mesonlib import Popen_safe, split_args
 
-def scanbuild(exename, srcdir, blddir, privdir, logdir, args):
+
+def scanbuild(exelist, srcdir, blddir, privdir, logdir, args):
     with tempfile.TemporaryDirectory(dir=privdir) as scandir:
-        meson_cmd = [exename] + args
-        build_cmd = [exename, '-o', logdir, 'ninja']
+        meson_cmd = exelist + args
+        build_cmd = exelist + ['-o', logdir, detect_ninja(), '-C', scandir]
         rc = subprocess.call(meson_cmd + [srcdir, scandir])
         if rc != 0:
             return rc
         return subprocess.call(build_cmd)
 
+
 def run(args):
     srcdir = args[0]
     blddir = args[1]
@@ -33,8 +37,10 @@ def run(args):
     privdir = os.path.join(blddir, 'meson-private')
     logdir = os.path.join(blddir, 'meson-logs/scanbuild')
     shutil.rmtree(logdir, ignore_errors=True)
-    exename = os.environ.get('SCANBUILD', 'scan-build')
-    if not shutil.which(exename):
-        print('Scan-build not installed.')
+
+    exelist = detect_scanbuild()
+    if not exelist:
+        print('Could not execute scan-build "%s"' % ' '.join(exelist))
         return 1
-    return scanbuild(exename, srcdir, blddir, privdir, logdir, meson_cmd)
+
+    return scanbuild(exelist, srcdir, blddir, privdir, logdir, meson_cmd)
